from copy import deepcopy

from gbvision.models.shapes import rect_collision
from gbvision.constants.types import Rect
from .continues_shape import ContinuesShape
from gbvision.utils.shapes.base_rect import BaseRect


class ContinuesRect(ContinuesShape):
    """
    An implementation of ContinuesShape to rectangles.
    used to try and check whether two rectangles are indeed the same one.
    """

    @staticmethod
    def _base_shape():
        return BaseRect

    def __init__(self, shape: Rect, *args, **kwargs):
        ContinuesShape.__init__(self, shape=shape, *args, **kwargs)

    def _shape_collision(self, shape) -> bool:
        return rect_collision(self._shape, shape)

    @staticmethod
    def _from_bounding_rect(bounding_rect: Rect):
        return deepcopy(bounding_rect)

    @staticmethod
    def _to_bounding_rect(rect: Rect) -> Rect:
        return deepcopy(rect)



import cv2

from gbvision.utils.pipeline import PipeLine


class MedianBlur(PipeLine):
    """
    creates a pipeline that blurs the given frame using the median blur method
    works very good for denoising purposes

    :param ksize: the size of the kernel used by the filter, must be an odd number
    :return: a pipeline that filters the image using the median blur method
    """

    def __init__(self, ksize: int):
        PipeLine.__init__(self, lambda frame: cv2.medianBlur(frame, ksize))



import cv2

from .base_shape import BaseShape
from gbvision.constants.types import Polygon, Number, Point
from gbvision.models.contours import contour_center


class BasePolygon(BaseShape):
    @staticmethod
    def shape_area(shape: Polygon) -> Number:
        return cv2.contourArea(shape)

    @staticmethod
    def shape_center(shape: Polygon) -> Point:
        return contour_center(shape)

    @staticmethod
    def shape_collision(shape1: Polygon, shape2: Polygon) -> bool:
        return NotImplemented




from .continues_circle import ContinuesCircle
from .continues_rect import ContinuesRect
from .continues_rotated_rect import ContinuesRotatedRect
from .continues_shape import ContinuesShape
from .continues_shape_wrapper import ContinuesShapeWrapper



import cv2
import numpy as np

from gbvision.utils.pipeline import PipeLine


@PipeLine
def corners(frame):
    """
    corner finding by a laplacian convolution of the frame by the matrix:
        [-1 1]
        [1 -1]

    :param frame: the frame to convolve
    :return: the convolved frame
    """
    return cv2.filter2D(frame, -1, np.array([[-1, 1], [1, -1]]))


@PipeLine
def edges(frame):
    """
    edges finding by a laplacian convolution of the frame by the matrix:
        [-1 -1 -1]
        [-1 8 -1]
        [-1 -1 -1]
    :param frame: the frame to convolve
    :return: the convolved frame
    """
    return cv2.filter2D(frame, -1, np.array([[-1, -1, -1], [-1, 8, -1], [-1, -1, -1]]))


@PipeLine
def sharpen(frame):
    """
    sharpens the frame by a laplacian convolution of the frame by the matrix:
        [-1 -1 -1]
        [-1 9 -1]
        [-1 -1 -1]

    :param frame: the frame to convolve
    :return: the convolved frame
    """
    return cv2.filter2D(frame, -1, np.array([[-1, -1, -1], [-1, 9, -1], [-1, -1, -1]]))


@PipeLine
def blur(frame):
    """
    blurs the frame by a laplacian convolution of the frame by the matrix:
        [1/9 1/9 1/9]
        [1/9 1/9 1/9]
        [1/9 1/9 1/9]

    :param frame: the frame to convolve
    :return: the convolved frame
    """
    return cv2.filter2D(frame, -1, np.array([[1, 1, 1], [1, 1, 1], [1, 1, 1]]) / 9)


@PipeLine
def blue(frame):
    """
    gets the blue channel of the frame

    :param frame: the frame
    :return: the blue channel only (as a grayscale frame)
    """
    return frame[:, :, 0]


@PipeLine
def green(frame):
    """
    gets the green channel of the frame

    :param frame: the frame
    :return: the green channel only (as a grayscale frame)
    """
    return frame[:, :, 1]


@PipeLine
def red(frame):
    """
    gets the red channel of the frame

    :param frame: the frame
    :return: the red channel only (as a grayscale frame)
    """
    return frame[:, :, 2]


@PipeLine
def gray(frame):
    """
    turns the frame to grayscale

    :param frame: the frame
    :return: the frame in grayscale form
    """
    return cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)


@PipeLine
def normalize(frame):
    """
    normalizes the frame to a pixel range of 0-1
    equivalent to (frame - min(frame)) / max(abs(frame - min(frame)))

    :param frame: the frame
    :return: the normalized frame (data type float32)
    """
    return cv2.normalize(frame, None, alpha=0.0, beta=1.0, norm_type=cv2.NORM_MINMAX, dtype=cv2.CV_32F)


@PipeLine
def distance_transform(frame):
    """
    performs the distance transform algorithm on a binary frame

    :param frame: the frame (binary, usually after threshold)
    :return: the distance transform of the frame using euclidean distance method
    """
    return cv2.distanceTransform(frame, cv2.DIST_L2, 3)


normalized_distance_transform = distance_transform + normalize



from .camera import CameraData, Camera
from gbvision.models.cameras import UNKNOWN_CAMERA
import cv2
import platform
import sys
import subprocess


class USBCamera(cv2.VideoCapture, Camera):
    """
    a basic usb connected camera which inherits from cv2 VideoCapture

    :param port: the usb port to which the camera is connected
    :param data: the camera data object that describes this camera
    """

    def __init__(self, port: int, data: CameraData = UNKNOWN_CAMERA):
        """
        initializes the camera
        
        """
        cv2.VideoCapture.__init__(self, port)
        self._data = data.copy()
        self.port = port

    def is_opened(self) -> bool:
        return self.isOpened()

    @staticmethod
    def __is_on_linux() -> bool:
        return platform.system() == 'Linux'

    def __v4l2_ctl_command(self, cmd, value) -> int:
        try:
            return subprocess.call(['v4l2-ctl', '-d', f'/dev/video{self.port}', '-c', f'{cmd}={value}'])
        except FileNotFoundError:
            print(
                "[WARN] setting some parameters such as exposure may not on a Linux machine work if you do not have "
                "v4l2 installed, if the command you tried to run does not work please install v4l2 using 'sudo apt "
                "install v4l-utils'",
                file=sys.stderr)
            return -1

    def set_exposure(self, exposure) -> bool:
        if self.__is_on_linux():
            if type(exposure) is bool:
                _exposure = int(exposure) + 10
            else:
                _exposure = exposure
            code = self.__v4l2_ctl_command('exposure_absolute', _exposure)
            if code == 0:
                return True
        if type(exposure) is bool:
            return self.set(cv2.CAP_PROP_EXPOSURE, int(exposure))
        return self.set(cv2.CAP_PROP_EXPOSURE, exposure)

    def set_auto_exposure(self, auto) -> bool:
        if self.__is_on_linux():
            if type(auto) is bool:
                _auto = 3 if auto else 1
            else:
                _auto = auto
            code = self.__v4l2_ctl_command('exposure_auto', _auto)
            if code == 0:
                return True
        if type(auto) is bool:
            return self.set(cv2.CAP_PROP_AUTO_EXPOSURE, 0.75 if auto else 0.25)
        return self.set(cv2.CAP_PROP_AUTO_EXPOSURE, auto)

    def get_data(self):
        return self._data

    def get_width(self):
        return self.get(cv2.CAP_PROP_FRAME_WIDTH)

    def get_height(self):
        return self.get(cv2.CAP_PROP_FRAME_HEIGHT)

    def _set_width(self, width):
        return self.set(cv2.CAP_PROP_FRAME_WIDTH, width)

    def _set_height(self, height):
        return self.set(cv2.CAP_PROP_FRAME_HEIGHT, height)

    def get_fps(self):
        return self.get(cv2.CAP_PROP_FPS)

    def set_fps(self, fps):
        return self.set(cv2.CAP_PROP_FPS, fps)



from typing import List, Tuple, Union

import cv2

from gbvision.utils.thresholds.threshold import Threshold
from gbvision.constants.types import Number


def bgr_threshold(frame, params):
    red, green, blue = params
    return cv2.inRange(frame, (red[0], green[0], blue[0]), (red[1], green[1], blue[1]))


class __ColorThresholdFunction:
    def __init__(self, color_type):
        self.color_type = color_type

    def __call__(self, frame, params):
        frame = cv2.cvtColor(frame, self.color_type)
        return bgr_threshold(frame, params)


hls_threshold = __ColorThresholdFunction(cv2.COLOR_BGR2HLS)
hsv_threshold = __ColorThresholdFunction(cv2.COLOR_BGR2HSV)
rgb_threshold = __ColorThresholdFunction(cv2.COLOR_BGR2RGB)
luv_threshold = __ColorThresholdFunction(cv2.COLOR_BGR2LUV)
lab_threshold = __ColorThresholdFunction(cv2.COLOR_BGR2LAB)
yuv_threshold = __ColorThresholdFunction(cv2.COLOR_BGR2YUV)
xyz_threshold = __ColorThresholdFunction(cv2.COLOR_BGR2XYZ)


def gray_threshold(frame, params):
    if len(frame.shape) > 2:
        frame = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
    return cv2.threshold(frame, params[0][0], params[0][1], cv2.THRESH_BINARY)[1]


_THRESHOLD_NAME_TABLE = {
    'BGR': bgr_threshold,
    'RGB': rgb_threshold,
    'HLS': hls_threshold,
    'HSV': hsv_threshold,
    'LUV': luv_threshold,
    'LAB': lab_threshold,
    'YUV': yuv_threshold,
    'XYZ': xyz_threshold,
    'GRAY': gray_threshold
}


class ColorThreshold(Threshold):
    """
    a class that represents a function that maps from an image to a binary image
    where 255 states that the pixel at the original image was in a range represented by the threshold object
    and 0 states the pixel was out of the range
    for example:
    Threshold([[200, 255], [0, 50], [0, 50]], 'RGB')
    the above threshold represents a relatively red pixel. when an image is filtered by it, every pixel that is
    relatively red will be given the value 255, and every pixel that isn't will be given the value of 0


    :param pixel_range: the threshold parameters, as a list of integers
        in the shape of 3x2 for a color image and 1x2 for a gray image
    :param thresh_type: a string, the type of color encoding to transform the image before applying the range test \
        binary filter \
        can be selected from the given list: \
        'BGR': default opencv color encoding (no change) \
        'RGB': default opencv color encoding in reverse \
        'HLS': hue, luminous, saturation \
        'HSV': hue, saturation, value \
        'LUV': https://en.wikipedia.org/wiki/CIELUV \
        'LAB': https://en.wikipedia.org/wiki/CIELAB_color_space \
        'YUV': https://en.wikipedia.org/wiki/YUV \
        'XYZ': https://en.wikipedia.org/wiki/CIE_1931_color_space \
        'GRAY': grayscale images (single channel), image presented doesn't have to be gray, the threshold will convert it
    """

    THRESH_TYPE_BGR = 'BGR'
    THRESH_TYPE_RGB = 'RGB'
    THRESH_TYPE_HLS = 'HLS'
    THRESH_TYPE_HSV = 'HSV'
    THRESH_TYPE_LUV = 'LUV'
    THRESH_TYPE_LAB = 'LAB'
    THRESH_TYPE_YUV = 'YUV'
    THRESH_TYPE_XYZ = 'XYZ'
    THRESH_TYPE_GRAY = 'GRAY'

    def __init__(self, pixel_range: List[Union[Tuple[Number, Number], List[Number]]], thresh_type='HSV'):
        assert thresh_type.upper() in _THRESHOLD_NAME_TABLE
        self.params = pixel_range
        self.type = thresh_type.upper()

    def __len__(self):
        """
        the amount of parameters is equal to the amount of channels

        :return: 1 if the threshold is for grayscale images, 3 if it is for color images
        """
        return len(self.params)

    def __getitem__(self, item: int):
        """
        returns the item'th channel's range

        :param item: the index
        :return: the range of the pixel in the item'th channel
        """
        return self.params[item]

    def __setitem__(self, key: int, value: list or tuple):
        """
        sets the key'th channel's range to value

        :param key: the channel
        :param value: an array of 2 integers, the lower and upper bounds of the pixel
        """
        self.params[key] = value

    def __iter__(self):
        """
        :return: an iterator that iterates through the channel ranges
        """
        return iter(self.params)

    def _threshold(self, frame):
        """
        activates the threshold filter on the given image
        :param frame: the image to activate the threshold on
        :return: a binary image, the frame after the threshold filter
        """
        return _THRESHOLD_NAME_TABLE[self.type](frame, self.params)

    def __repr__(self):
        return f'<{self}>'

    def __str__(self):
        return f"ColorThreshold({self.params}, '{self.type}')"



import socket

from gbvision.constants.net import LOCAL_SERVER_IP
from .stream_broadcaster import StreamBroadcaster
from gbvision.exceptions.tcp_stream_closed import TCPStreamClosed


class TCPStreamBroadcaster(StreamBroadcaster):
    """
    this class uses TCP to send a stream over the network, the stream is by default set to be MJPEG
    the broadcaster is the server and the receiver is the client
    
    :param port: the port which TCP will be using
    :param im_encode: the type of image encoding to send over the network, default is .jpg (JPEG)
        for missing parameters, see documentation on StreamBroadcaster
    """

    def __init__(self, port: int, *args, **kwargs):
        StreamBroadcaster.__init__(self, *args, **kwargs)
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.server_addr = (LOCAL_SERVER_IP, port)
        self.socket.bind(self.server_addr)
        self.socket.listen(10)
        self.socket, addr = self.socket.accept()

    def _send_frame(self, frame):
        try:
            self.socket.send(frame)
        except IOError as e:
            raise TCPStreamClosed() from e
        self._update_time()



from typing import Union, Tuple

from .erode import Erode
from .dilate import Dilate

from gbvision.utils.pipeline import PipeLine


class ErodeAndDilate(PipeLine):
    """
    a pipeline class that erodes and dilates the given frame by the same kernel

    :param ksize: the kernel size, either an integer (meaning an nxn kernel) or a tuple (nxm kernel)
    :param iterations: optional, the amount of Dilate iterations to perform, default is 1.
            Note! a large number of iterations will slow down the program
    """

    def __init__(self, ksize: Union[int, Tuple[int, int]], iterations=1):
        PipeLine.__init__(self)
        self.functions += (Erode(ksize, iterations) + Dilate(ksize, iterations)).functions



from typing import List

from gbvision.constants.types import Frame, RotatedRect, FilterFunction

from gbvision.models.system import EMPTY_PIPELINE
from gbvision.models.contours import find_contours, FilterContours, contours_to_rotated_rects
from .object_finder import ObjectFinder
from gbvision.utils.shapes.base_rotated_rect import BaseRotatedRect


class RotatedRectFinder(ObjectFinder):
    """
    finds a rectangular object, but rotated. recommended to use when you know the shape isn't parallel to the camera

    :param threshold_func: a pipeline (or any sort of function) that returns a binary threshold of the object
     the finder is searching, the object needs to be white and the rest if the image black (doesn't
     have to be perfect)
    :param area_scalar: optional, a scalar to multiply the area by, for fine tuning of the function's output
    :param contour_min_area: the minimal area of a contour, used for FilterContours, default is 0 (no area limit)
    :param contours_process: a pipeline to run on the list of contours (optional)
    :param rotated_rects_process: a pipeline to run on the list of rotated rects (optional)
    """

    @staticmethod
    def _base_shape():
        return BaseRotatedRect

    def __init__(self, threshold_func: FilterFunction, game_object, area_scalar=1.0, contour_min_area=0.0,
                 contours_process=EMPTY_PIPELINE, rotated_rects_process=EMPTY_PIPELINE):
        ObjectFinder.__init__(self, game_object, area_scalar=area_scalar)
        self._full_pipeline = (EMPTY_PIPELINE +
                               threshold_func +
                               find_contours +
                               FilterContours(min_area=contour_min_area) +
                               contours_process +
                               contours_to_rotated_rects +
                               rotated_rects_process)

    def find_shapes_unsorted(self, frame: Frame) -> List[RotatedRect]:
        return self._full_pipeline(frame)



from .readable_window import ReadableWindow


class StreamWindow(ReadableWindow):
    """
    a basic window that displays the stream from a stream receiver
    
    :type self.wrap_object: gbvision.StreamReceiver
    """



import cv2

from .base_shape import BaseShape
from gbvision.constants.types import Contour, Number, Point
from gbvision.models.contours import contour_center


class BaseContour(BaseShape):
    @staticmethod
    def shape_area(shape: Contour) -> Number:
        return cv2.contourArea(shape)

    @staticmethod
    def shape_center(shape: Contour) -> Point:
        return contour_center(shape)

    @staticmethod
    def shape_collision(shape1: Contour, shape2: Contour) -> bool:
        return NotImplemented




import abc

from .window import Window
from gbvision.models.system import EMPTY_PIPELINE
from gbvision.utils.recorders.recorder import Recorder


class RecordingWindow(Window, abc.ABC):
    """
    A basic window that records the stream it receives

    :param recording_pipeline: a drawing pipeline to run on the recorded frame, usually you will want this to be the
        same as the drawing pipeline
    """

    def __init__(self, window_name: str, recorder: Recorder, drawing_pipeline=EMPTY_PIPELINE,
                 recording_pipeline=EMPTY_PIPELINE):
        Window.__init__(self, window_name=window_name, drawing_pipeline=drawing_pipeline)
        self.recording_pipeline = recording_pipeline
        self.recorder = recorder

    def show_frame(self, frame):
        self.recorder.record(self.recording_pipeline(frame))
        return Window.show_frame(self, frame)

    def _close(self):
        self.recorder.close()



from gbvision.utils.thresholds.color_threshold import ColorThreshold
from gbvision.utils.pipeline import PipeLine


class DistanceTransformThreshold(PipeLine):
    """
    a pipeline factory that performs normalized distance transform and then a minimum threshold on the frame which
    removes from the original frame all white pixels that are at most min_distance_ratio normalized distance from the
    nearest black frame

    :param min_distance_ratio: the minimum ratio between the maximum distance of a pixel from a white pixel in the
        frame and a certain pixel for it to be included in the threshold, between 0 and 1
    """
    def __init__(self, min_distance_ratio: float):
        PipeLine.__init__(self)
        from gbvision.models.basic_ops import normalized_distance_transform
        self.functions += (normalized_distance_transform + ColorThreshold([[min_distance_ratio, 1.0]],
                                                                          ColorThreshold.THRESH_TYPE_GRAY)).functions



from .base_shape import BaseShape
from gbvision.constants.types import RotatedRect, Number, Point
from gbvision.models.shapes import rotated_rect_collision


class BaseRotatedRect(BaseShape):
    @staticmethod
    def shape_center(shape: RotatedRect) -> Point:
        return shape[0]

    @staticmethod
    def shape_collision(shape1: RotatedRect, shape2: RotatedRect) -> bool:
        return rotated_rect_collision(shape1, shape2)

    @classmethod
    def shape_area(cls, shape: RotatedRect) -> Number:
        return shape[1][0] * shape[1][1]



from gbvision.utils.net import AsyncStreamReceiver
from .tcp_stream_receiver import TCPStreamReceiver


class AsyncTCPStreamReceiver(AsyncStreamReceiver, TCPStreamReceiver):

    def __init__(self, ip, port, *args, **kwargs):
        TCPStreamReceiver.__init__(self, ip, port, *args, **kwargs)
        AsyncStreamReceiver.__init__(self, *args, **kwargs)

    def _read(self):
        return TCPStreamReceiver.read(self)



import abc
from typing import Type, List

import numpy as np

from gbvision.constants.types import Shape, Number, Rect, Point


class BaseShape(abc.ABC):
    @staticmethod
    @abc.abstractmethod
    def shape_collision(shape1: Shape, shape2: Shape) -> bool:
        """
        checks if the two shapes are colliding

        :param shape1: the first shape
        :param shape2: the second shape
        :return: True if the shapes are colliding, False otherwise
        """

    @staticmethod
    @abc.abstractmethod
    def shape_area(shape: Shape) -> Number:
        """
        calculates the area of the shape

        :param shape: the shape
        :return: the area of the shape
        """

    @staticmethod
    @abc.abstractmethod
    def shape_center(shape: Shape) -> Point:
        """
        calculates the center-of-mass of the shape

        :param shape: the shape
        :return: the center of the shape
        """

    @classmethod
    def shape_root_area(cls, shape: Shape) -> Number:
        """
        calculates the square root of the area of the shape
        default is the square root of cls.shape_area, but it can be overridden in case there is a simpler way\
         (for example for circles)

        :param shape: the shape
        :return: the square root of the area of the shape
        """
        return np.sqrt(cls.shape_area(shape))

    @classmethod
    def sort_shapes(cls, shapes: List[Shape]) -> List[Shape]:
        """
        sorts the list of shapes by area, should be overridden to use area root in case it's better
        doesn't modify the given list, returns a new list

        :param shapes: the list of shapes to sort
        :return: a sorted copy of the list of shapes
        """
        return sorted(shapes, key=cls.shape_area, reverse=True)

    @classmethod
    def filter_inner_shapes(cls, shapes: List[Shape]) -> List[Shape]:
        """
        filters out all shapes that are colliding with a shape with a smaller index in the given list
        returns a list of all shapes that aren't colliding

        :param shapes: the list of shapes
        :return: the filtered list of shapes
        """
        filtered_shapes = []
        for i, shape in enumerate(shapes):
            shape_invalid = False
            for j in range(i):
                shape_invalid = cls.shape_collision(shape, shapes[j])
                if shape_invalid:
                    break
            if not shape_invalid:
                filtered_shapes.append(shape)
        return filtered_shapes


BaseShapeType = Type[BaseShape]



from typing import Union, Tuple
import cv2
import numpy as np

from gbvision.utils.pipeline import PipeLine


class Dilate(PipeLine):
    """
        creates a pipeline that dilates the image by a kernel of ones
        used mainly for Erode & Dilate denoise filters

        :param ksize: the kernel size, either an integer (meaning an nxn kernel) or a tuple (nxm kernel)
        :param iterations: optional, the amount of Dilate iterations to perform, default is 1.
            Note! a large number of iterations will slow down the program
        :return: a pipeline that dilates the given frame
        """

    def __init__(self, ksize: Union[int, Tuple[int, int]], iterations=1):
        if type(ksize) is int:
            ksize = (ksize, ksize)
        PipeLine.__init__(self, lambda frame: cv2.dilate(frame, np.ones(ksize), iterations=iterations))



from abc import ABC

import cv2

from gbvision.models.system import EMPTY_PIPELINE
from .opencv_window import OpenCVWindow
from .wrapper_window import WrapperWindow


class WrapperOpenCVWindow(OpenCVWindow, WrapperWindow, ABC):
    """
    a basic window that displays a feed from a camera

    :param wrap_object: the object from which to read the frame
    :param drawing_pipeline: optional, a pipeline of drawing functions that will run on the frame before displaying
        it
    :param exit_button: an array of keys (a string), when one of the keys are pressed the window will be closed
    :param window_name: the title of the window
    """

    def __init__(self, window_name: str, wrap_object, exit_button='qQ', drawing_pipeline=EMPTY_PIPELINE,
                 flags=cv2.WINDOW_FREERATIO):
        """
        initializes the stream window
        
        """
        OpenCVWindow.__init__(self, window_name, flags=flags, exit_button=exit_button)
        WrapperWindow.__init__(self, window_name=window_name, wrap_object=wrap_object,
                               drawing_pipeline=drawing_pipeline)



from .async_readable import AsyncReadable
from .game_object import GameObject
from .pipeline import PipeLine
from .readable import Readable
from .tracker import Tracker
from .cameras import *
from .continuity import *
from .finders import *
from .net import *
from .thresholds import *
from .shapes import *
from .recorders import *
from .denoising import *



from .finding_tools import viewing_angle_of_object, plane_distance_from_object, plane_angle_by_location, \
    distance_from_object
from .image_tools import crop, median_threshold
from .list_tools import split_list



from .could_not_read_frame_exception import CouldNotReadFrameException
from .device_not_found_exception import DeviceNotFoundException
from .stream_closed import StreamClosed
from .tcp_stream_closed import TCPStreamClosed
from .vision_exception import VisionException
from .vision_warning import VisionWarning



from typing import List

import cv2
import numpy as np

from gbvision.constants.types import Frame, Contour, Color, Circle, Rect, RotatedRect, Ellipse, Coordinates, Number, \
    Line


def draw_contours(frame: Frame, cnts: List[Contour], color: Color, *args, **kwargs) -> Frame:
    """
    draws all contours on a copy of the frame and returns the copy

    :param frame: the frame to draw on
    :param cnts: the contours to draw
    :param color: the color to draw in
    :param args: all extra args to opencv's  drawContours (for example thickness)
    :param kwargs: all extra keyword args to opencv's  drawContours (for example thickness)
    :return: a copy of the frame, after drawing
    """
    frame = frame.copy()
    cv2.drawContours(frame, cnts, -1, color, *args, **kwargs)
    return frame


def draw_circles(frame: Frame, circs: List[Circle], color: Color, *args, **kwargs) -> Frame:
    """
    draws all circles on a copy of the frame and returns the copy

    :param frame: the frame to draw on
    :param circs: the circles to draw
    :param color: the color to draw in
    :param args: all extra args to opencv's circle (for example thickness)
    :param kwargs: all extra keyword args to opencv's circle (for example thickness)
    :return: a copy of the frame, after drawing
    """
    frame = frame.copy()
    for c in circs:
        cv2.circle(frame, (int(c[0][0]), int(c[0][1])), int(c[1]), color, *args, **kwargs)
    return frame


def draw_rects(frame: Frame, rects: List[Rect], color: Color, *args, **kwargs) -> Frame:
    """
    draws all rects on a copy of the frame and returns the copy

    :param frame: the frame to draw on
    :param rects: the rects to draw
    :param color: the color to draw in
    :param args: all extra args to opencv's rectangle (for example thickness)
    :param kwargs: all extra keyword args to opencv's rectangle (for example thickness)
    :return: a copy of the frame, after drawing
    """
    frame = frame.copy()
    for r in rects:
        cv2.rectangle(frame, (int(r[0]), int(r[1])), (int(r[0] + r[2]), int(r[1] + r[3])), color, *args, **kwargs)
    return frame


def draw_rotated_rects(frame: Frame, rotated_rects: List[RotatedRect], color: Color, *args, **kwargs) -> Frame:
    """
    draws all rotated rects on a copy of the frame and returns the copy

    :param frame: the frame to draw on
    :param rotated_rects: the rotated rects to draw
    :param color: the color to draw in
    :param args: all extra args to opencv's drawContours (for example thickness)
    :param kwargs: all extra keyword args to opencv's drawContours (for example thickness)
    :return: a copy of the frame, after drawing
    """
    frame = frame.copy()
    for r in rotated_rects:
        box = cv2.boxPoints(r)
        box = np.int0(box)
        cv2.drawContours(frame, [box], 0, color, *args, **kwargs)
    return frame


def draw_ellipses(frame: Frame, ellipses: List[Ellipse], color: Color, *args, **kwargs) -> Frame:
    """
    draws all contours on a copy of the frame and returns the copy
    
    :param frame: the frame to draw on
    :param ellipses: the ellipses to draw
    :param color: the color to draw in
    :param args: all extra args to opencv's ellipse (for example thickness)
    :param kwargs: all extra keyword args to opencv's ellipse (for example thickness)
    :return: a copy of the frame, after drawing
    """
    frame = frame.copy()
    for e in ellipses:
        cv2.ellipse(frame, e, color, *args, **kwargs)
    return frame


def draw_lines(frame: Frame, lines: List[Line], color: Color, *args, **kwargs) -> Frame:
    """
    draws all Lines on a copy of the frame and returns the copy

    :param frame: the frame to draw on
    :param lines: the list of lines to draw
    :param color: the color to draw in
    :param args: all extra args to opencv's lines (for example thickness)
    :param kwargs: all extra keyword args to opencv's lines (for example thickness)
    :return: a copy of the frame, after drawing
    """

    frame = frame.copy()
    for line in lines:
        cv2.line(frame, line[0], line[1], color, *args, **kwargs)
    return frame


def draw_text(frame: Frame, text: str, coords: Coordinates, font_scale: Number, color: Color,
              font=cv2.FONT_HERSHEY_SIMPLEX, *args, **kwargs) -> Frame:
    """
    draws the text on a copy of the frame and returns the copy

    :param frame: the frame to draw on
    :param text: the text to draw
    :param coords: the coordinates of the bottom-left corner of the text
    :param font_scale: the size of the drawn text (multiplied by the default size of the font)
    :param color: the color to draw the text in
    :param font: the font, an opencv font constant, default is cv2.FONT_HERSHEY_SIMPLEX
    :param args: additional arguments to cv2.putText
    :param kwargs: additional keyword arguments to cv2.putText (such as thickness)
    :return: a copy of the frame with the text drawn on it
    """
    frame = frame.copy()
    cv2.putText(frame, text, coords, font, font_scale, color, *args, **kwargs)
    return frame



from .base_shape import BaseShape
from gbvision.constants.types import Rect, Number, Point
from gbvision.models.shapes import rect_collision


class BaseRect(BaseShape):
    @staticmethod
    def shape_center(shape: Rect) -> Point:
        return shape[0] + (shape[2] / 2), shape[1] + (shape[3] / 2)

    @staticmethod
    def shape_collision(shape1: Rect, shape2: Rect) -> bool:
        return rect_collision(shape1, shape2)

    @classmethod
    def shape_area(cls, shape: Rect) -> Number:
        return shape[2] * shape[3]



from typing import Any, Callable
import functools


class PipeLine:
    """
    a class representing a pipeline of function
    each function receives one input, which is the output of the previous function in the pipeline
    pipelines are great for representing a long computer vision function (which is why such functions
    are called pipelines).
    the PipeLine class can also be used as a function decorator
    for example, creating a pipeline that adds 1 to it's output can be done in two ways:
    Example::
        inc = PipeLine(lambda x: x + 1)
        three = inc(2)
        
    or
    Example::
        @PipeLine
        def inc(x):
            return x + 1
        three = inc(2)


    You can create a PipeLine from multiple functions:
    Example::
        open_and_read_file = PipeLine(open, lambda x: x.read())
        text = open_and_read_file("file.txt")
            
    You can also inherit from the PipeLine class to make a PipeLine factory:
    Example::
        class Adder(PipeLine):
            def __init__(self, num):
                self.num = num
                PipeLine.__init__(self, self.adding_func)
            def adding_func(self, item):
                return item + self.num

    You can also do it like this:
    Example::
        class Adder(PipeLine):
            def __init__(self, num):
                def adding_func(item):
                    return item + num

                PipeLine.__init__(self, adding_func)

    You can use combine a few PipeLines together to create function composition:
    Example::
        multiply_by_2_then_add_3 = PipeLine(lambda x: x * 2) + PipeLine(lambda x: x + 3)

    :param functions: a tuple of functions to run one after the other as a pipeline
    """

    def __init__(self, *functions: Callable[[Any], Any]):
        self.functions = list(functions)

    def __call__(self, image):
        """
        activate this pipeline and return the result

        :param image: the input to the first function in the pipeline (also the input to the entire pipeline) \
        doesn't have to be an image, can be anything
        :return: the output of the last function in the pipeline, can be data type
        """
        return functools.reduce(lambda x, f: f(x), self.functions, image)

    def __add__(self, other):
        """
        creates a new pipeline which uses the output of this pipeline as input to the other pipeline

        :param other: the second pipeline
        :return: a new pipeline, and calling this pipeline with the parameter image is similar to \
        performing other(self(image))
        """
        if isinstance(other, PipeLine):
            return PipeLine(*self.functions + other.functions)
        return PipeLine(*self.functions + [other])

    def __radd__(self, other):
        """
        adds this PipeLine to another function that isn't a PipeLine

        :param other: the function
        :return: a new PipeLine which performs self(other(image)) on the parameter image
        """
        return PipeLine(other) + self

    def __getitem__(self, item):
        """
        gets the function at index item

        :param item: the index
        :return: the item
        """
        return self.functions[item]

    def __setitem__(self, key, value):
        """
        sets the function at index key to the new function value

        :param key: the index
        :param value: the new function
        """
        self.functions[key] = value

    def __iter__(self):
        """
        :return: an iterator that iterates through all functions in this pipeline
        """
        return iter(self.functions)



import abc
from typing import Union

import numpy as np
from gbvision.constants.types import Number

from .camera_data import CameraData
from gbvision.utils.readable import Readable


class Camera(Readable, abc.ABC):
    """
    an abstract class representing a camera
    """

    @abc.abstractmethod
    def release(self):
        """
        closes the handle to this camera, if it is not necessary please override this method to a blank method
        """

    @abc.abstractmethod
    def is_opened(self) -> bool:
        """
        checks if the camera can be read from

        :return: True if the camera can be read from, otherwise False
        """

    @abc.abstractmethod
    def set_exposure(self, exposure: Union[int, float, bool]) -> bool:
        """
        sets the camera's exposure

        :param exposure: the new exposure
        :return: True on success, False on failure
        """

    @abc.abstractmethod
    def set_auto_exposure(self, auto: Union[int, float, bool]) -> bool:
        """
        sets the camera's auto exposure

        :param auto: the new auto exposure
        :return: True on success, False on failure
        """

    @abc.abstractmethod
    def get_data(self) -> CameraData:
        """
        :return: this camera's constant descriptor (must be the real descriptor, can't be a copy) \
            when the values of this descriptor are changed, the values of the real camera descriptor must also change
        """

    @abc.abstractmethod
    def get_width(self) -> int:
        """
        :return: the width of a frame read from this camera
        """

    @abc.abstractmethod
    def get_height(self) -> int:
        """
        :return: the height of a frame read from this camera
        """

    @abc.abstractmethod
    def _set_width(self, width: int):
        """
        unsafe set width
        supposed to be overridden and only used by rescale, resize and set_frame size methods
        never to be used by the programmer, only by the api

        :param width: new width
        """

    @abc.abstractmethod
    def _set_height(self, height: int):
        """
        unsafe set height
        supposed to be overridden and only used by rescale, resize and set_frame size methods
        never to be used by the programmer, only by the api

        :param height: new height
        """

    @abc.abstractmethod
    def get_fps(self) -> Number:
        """
        gets the fps of this camera

        :return: the fps of the camera
        """

    @abc.abstractmethod
    def set_fps(self, fps: Number) -> bool:
        """
        sets the fps for this camera

        :return: True on success, False otherwise
        """

    def rescale(self, factor: float):
        """
        rescale the size of the frames read from this camera by a factor

        :param factor: the rescaling factor
        """
        self._set_width(int(self.get_width() * factor))
        self._set_height(int(self.get_height() * factor))
        self.get_data().focal_length *= factor

    def resize(self, fx: float, fy: float):
        """
        rescale the size of the frames read from this camera by different width and height factors

        :param fx: the width factor
        :param fy: the height factor
        """
        self._set_width(int(self.get_width() * fx))
        self._set_height(int(self.get_height() * fy))
        self.get_data().focal_length *= np.sqrt(fx * fy)

    def set_frame_size(self, width: int, height: int):
        """
        reset the width and height of frames read from this camera to given values
        
        :param width: the new width
        :param height: the new height
        """
        old_width, old_height = self.get_width(), self.get_height()
        self._set_height(height)
        self._set_width(width)
        self.get_data().focal_length *= np.sqrt(width * height / (old_width * old_height))



# constants
from .constants import *

# exceptions
from .exceptions import *

# gui
from .gui import *

# models
from .models import *

# tools
from .tools import *

# utils
from .utils import *


# configure opencv
def cv_config():
    """
    configure some opencv stuff so that it doesn't cause problems
    called automatically when importing gbvision, and does not need to be called by the user
    """
    import cv2
    if cv2.getVersionString()[0] == '2':
        for i in filter(lambda attr: attr.startswith("CV_CAP_PROP"), dir(cv2.cv)):
            object.__setattr__(cv2, i[3:], cv2.cv.__getattribute__(i))
            cv2.__dict__[i[3:]] = cv2.cv.__dict__[i]


cv_config()
del cv_config



from gbvision.utils.cameras.camera_data import CameraData

# LIFECAM_STUDIO = CameraData(648.5256168410046, 0.340394, 0, is_immutable=True, name='LIFECAM_STUDIO')

LIFECAM_3000 = CameraData(697.0395744431028, 0.4435563306578366, 0.3337068395920225, is_immutable=True,
                          name='LIFECAM_3000')

UNKNOWN_CAMERA = CameraData(0, 0, 0, is_immutable=True, name='UNKNOWN_CAMERA')



from gbvision.constants.types import Color, FilterFunction

from gbvision.models.system import EMPTY_PIPELINE
from gbvision.models.contours import find_contours, contours_to_circles, contours_to_rects, contours_to_rotated_rects, \
    contours_to_ellipses
from gbvision.utils.pipeline import PipeLine
from .drawing_functions import draw_contours, draw_circles, draw_rects, draw_rotated_rects, draw_ellipses


class _DrawObject(PipeLine):
    def __init__(self, finding_func, color, drawing_func, *args, **kwargs):
        def _draw(frame):
            return drawing_func(frame, finding_func(frame), color, *args, **kwargs)

        PipeLine.__init__(self, _draw)


class DrawContours(_DrawObject):
    """
    a pipeline that draws all contours according to the given parameters, and returns a copy of the frame after drawing
    """

    def __init__(self, threshold_func: FilterFunction, color: Color, contours_process=EMPTY_PIPELINE, *args,
                 **kwargs):
        contour_finding = EMPTY_PIPELINE + threshold_func + find_contours + contours_process
        _DrawObject.__init__(self, contour_finding, color, draw_contours, *args, **kwargs)


class DrawCircles(_DrawObject):
    """
    a pipeline that draws all circles according to the given parameters, and returns a copy of the frame after drawing
    """

    def __init__(self, threshold_func: FilterFunction, color: Color, contours_process=EMPTY_PIPELINE,
                 circle_process=EMPTY_PIPELINE, *args, **kwargs):
        circle_finding = EMPTY_PIPELINE + threshold_func + find_contours + contours_process + contours_to_circles + \
                         circle_process

        _DrawObject.__init__(self, circle_finding, color, draw_circles, *args, **kwargs)


class DrawRects(_DrawObject):
    """
    a pipeline that draws all rects according to the given parameters, and returns a copy of the frame after drawing
    """

    def __init__(self, threshold_func: FilterFunction, color: Color, contours_process=EMPTY_PIPELINE,
                 rects_process=EMPTY_PIPELINE, *args, **kwargs):
        rect_finding = EMPTY_PIPELINE + threshold_func + find_contours + contours_process + contours_to_rects + \
                       rects_process

        _DrawObject.__init__(self, rect_finding, color, draw_rects, *args, **kwargs)


class DrawRotatedRects(_DrawObject):
    """
    a pipeline that draws all rotated rects according to the given parameters, and returns a copy of the frame after drawing
    """

    def __init__(self, threshold_func: FilterFunction, color: Color, contours_process=EMPTY_PIPELINE,
                 rotated_rects_process=EMPTY_PIPELINE, *args, **kwargs):
        rotated_rect_finding = EMPTY_PIPELINE + threshold_func + find_contours + contours_process + contours_to_rotated_rects + \
                               rotated_rects_process

        _DrawObject.__init__(self, rotated_rect_finding, color, draw_rotated_rects, *args, **kwargs)


class DrawEllipses(_DrawObject):
    """
    a pipeline that draws all ellipses according to the given parameters, and returns a copy of the frame after drawing
    """

    def __init__(self, threshold_func: FilterFunction, color: Color, contours_process=EMPTY_PIPELINE,
                 ellipses_process=EMPTY_PIPELINE, *args, **kwargs):
        ellipses_finding = EMPTY_PIPELINE + threshold_func + find_contours + contours_process + contours_to_ellipses + \
                           ellipses_process

        _DrawObject.__init__(self, ellipses_finding, color, draw_ellipses, *args, **kwargs)



from typing import List

from gbvision.constants.math import SQRT_PI

from .base_shape import BaseShape
from gbvision.constants.types import Circle, Number, Point
from gbvision.models.shapes import circle_collision


class BaseCircle(BaseShape):
    @staticmethod
    def shape_center(shape: Circle) -> Point:
        return shape[0]

    @staticmethod
    def shape_collision(shape1: Circle, shape2: Circle) -> bool:
        return circle_collision(shape1, shape2)

    @classmethod
    def shape_area(cls, shape: Circle) -> Number:
        return cls.shape_root_area(shape) ** 2

    @classmethod
    def shape_root_area(cls, shape: Circle) -> Number:
        return shape[1] * SQRT_PI

    @classmethod
    def sort_shapes(cls, shapes: List[Circle]):
        return sorted(shapes, key=cls.shape_root_area)



from .base_shape import BaseShape
from .base_circle import BaseCircle
from .base_rect import BaseRect
from .base_convex_polygon import BaseConvexPolygon
from .base_contour import BaseContour
from .base_rotated_rect import BaseRotatedRect
from .base_polygon import BasePolygon



from .math import EPSILON, SQRT_PI
from .types import Frame, Coordinates, Location, RotatedRect, Shape, Number, Color, Circle, Contour, FilterFunction, \
    Rect, FixedPolygon, Ellipse, Polygon, ROI, Point, Line



import socket
from .stream_broadcaster import StreamBroadcaster


class UDPStreamBroadcaster(StreamBroadcaster):
    """
    this class uses UDP to send a stream over the network, the stream is by default set to be MJPEG
    the broadcaster is the client and the receiver is the server
    WARNING: do not use this class to send large images, udp has a limited packet size

    :param ip: the IPv4 address of the udp stream receiver, for example '10.45.90.5'
    :param port: the port that UDP uses to send packets on
    :param im_encode: the type of image encoding to send over the network, default is '.jpg' (JPEG)
    """

    def __init__(self, ip: str, port: int, *args, **kwargs):
        StreamBroadcaster.__init__(self, *args, **kwargs)
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.server_addr = (ip, port)

    def _send_frame(self, frame):
        self.socket.sendto(frame, self.server_addr)



from .recording_wrapper_opencv_window import RecordingWrapperOpenCVWindow
from .readable_window import ReadableWindow


class RecordingReadableWindow(RecordingWrapperOpenCVWindow, ReadableWindow):
    """
    a basic window that displays the video from a readable
    and records the video to a file
    """



from .recording_opencv_window import RecordingOpenCVWindow
from .feed_window import FeedWindow


class RecordingFeedWindow(RecordingOpenCVWindow, FeedWindow):
    """
    a basic window that displays the stream from a stream receiver
    """



from .camera_window import CameraWindow
from .drawing_functions import draw_circles, draw_contours, draw_ellipses, draw_rects, draw_rotated_rects, draw_text, \
    draw_lines
from .drawing_tools import DrawContours, DrawCircles, DrawEllipses, DrawRects, DrawRotatedRects
from .feed_window import FeedWindow
from .opencv_window import OpenCVWindow
from .readable_window import ReadableWindow
from .recording_camera_window import RecordingCameraWindow
from .recording_feed_window import RecordingFeedWindow
from .recording_opencv_window import RecordingOpenCVWindow
from .recording_readable_window import RecordingReadableWindow
from .recording_stream_window import RecordingStreamWindow
from .recording_window import RecordingWindow
from .recording_wrapper_opencv_window import RecordingWrapperOpenCVWindow
from .stream_window import StreamWindow
from .window import Window
from .wrapper_opencv_window import WrapperOpenCVWindow
from .wrapper_window import WrapperWindow



import abc

from gbvision.constants.types import Frame


class Recorder(abc.ABC):
    """
    An abstract recorder class
    records a given feed of frames into a file

    :param file_name: the file name
    """
    def __init__(self, file_name):
        self.file_name = file_name

    @abc.abstractmethod
    def record(self, frame: Frame):
        """
        records the frame

        :param frame: the frame to record
        """

    @abc.abstractmethod
    def close(self):
        """
        ends the writing to the file

        """

    @abc.abstractmethod
    def is_opened(self) -> bool:
        """
        checks if this video file is opened

        :return: True if this is opened, False otherwise
        """


from .recording_window import RecordingWindow
from .opencv_window import OpenCVWindow
from gbvision.models.system import EMPTY_PIPELINE
from os.path import splitext
import cv2
from gbvision.constants.video import VIDEO_FILE_TYPE
from gbvision.utils.recorders.opencv_recorder import OpenCVRecorder


class RecordingOpenCVWindow(OpenCVWindow, RecordingWindow):
    """
    A basic recording window that uses an OpenCVRecorder to record videos

    :param file_name: the name of the output file
    :param drawing_pipeline: optional, a pipeline of drawing functions that will run on the frame before displaying
        it
    :param exit_button: an array of keys (a string), when one of the keys are pressed the window will be closed
    :param window_name: the title of the window
    :param fps: the fps of the video file
    :param recording_pipeline: optional, a drawing pipeline to run on the frames being recorded
    """

    def __init__(self, window_name: str, file_name: str, fps=20.0, exit_button='qQ',
                 drawing_pipeline=EMPTY_PIPELINE, recording_pipeline=EMPTY_PIPELINE, width=None, height=None,
                 flags=cv2.WINDOW_FREERATIO):
        """
        initializes the stream window
        
        """

        recorder = OpenCVRecorder(file_name, fps, width, height)
        OpenCVWindow.__init__(self, window_name, exit_button=exit_button, flags=flags)
        RecordingWindow.__init__(self, window_name=window_name, recorder=recorder, drawing_pipeline=drawing_pipeline,
                                 recording_pipeline=recording_pipeline)

    def _close(self):
        OpenCVWindow._close(self)
        RecordingWindow._close(self)



from .base_polygon import BasePolygon
from gbvision.constants.types import Polygon
from gbvision.models.shapes import convex_shape_collision


class BaseConvexPolygon(BasePolygon):

    @staticmethod
    def shape_collision(shape1: Polygon, shape2: Polygon) -> bool:
        return convex_shape_collision(shape1, shape2)



from .recording_readable_window import RecordingReadableWindow
from .stream_window import StreamWindow


class RecordingStreamWindow(RecordingReadableWindow, StreamWindow):
    """
    a basic window that displays the stream from a stream receiver
    and records the stream to a video file
    """



from .readable_window import ReadableWindow


class CameraWindow(ReadableWindow):
    """
    a basic window that displays a feed from a camera
    in this class, self.wrap_object will be of type Camera
    
    :type self.wrap_object: gbvision.Camera
    """



from os.path import splitext
from typing import Optional

import cv2
from gbvision.constants.video import VIDEO_FILE_TYPE

from .recorder import Recorder


class OpenCVRecorder(Recorder):
    """
    a basic implementation of the recorder class using OpenCV

    :param file_name: the path to the output file
    :param fps: the fps of the video
    :param width: optional, the width of the video (will be set automatically if not given)
    :param height: optional, the height of the video (will be set automatically if not given)
    """
    def __init__(self, file_name, fps, width=None, height=None):
        Recorder.__init__(self, file_name)

        _, file_ext = splitext(file_name)

        self.fourcc = cv2.VideoWriter_fourcc(*VIDEO_FILE_TYPE[file_ext.upper()])

        self.fps = fps

        self.video_writer: Optional[cv2.VideoWriter] = None
        self.width = width
        self.height = height

    def record(self, frame):
        if frame is None or len(frame.shape) == 0:
            return
        if self.video_writer is None:
            self.width = self.width if self.width is not None else frame.shape[1]
            self.height = self.height if self.height is not None else frame.shape[0]
            self.video_writer = cv2.VideoWriter()
            self.video_writer.open(self.file_name, self.fourcc, self.fps, (self.width, self.height))
        self.video_writer.write(frame)

    def close(self):
        self.video_writer.release()

    def is_opened(self) -> bool:
        if self.video_writer is None:
            return True
        return self.video_writer.isOpened()



import abc
from typing import Any

from gbvision.constants.types import Frame
from .window import Window
from gbvision.models.system import EMPTY_PIPELINE
from threading import Thread


class WrapperWindow(Window, abc.ABC):
    """
    A window class that uses an object (wrap_object) and reads frames from it in to display it's feed

    :param wrap_object: an object to read frames from, can be of any type
    """

    def __init__(self, window_name: str, wrap_object: Any, drawing_pipeline=EMPTY_PIPELINE):
        Window.__init__(self, window_name=window_name, drawing_pipeline=drawing_pipeline)
        self.wrap_object = wrap_object

    def show_and_get_frame(self) -> Frame:
        """
        shows one frame and returns it

        :return: the frame if the window was not closed, None otherwise
        """
        frame = self._get_frame()
        if self.show_frame(frame):
            return frame
        return None

    def show(self):
        """
        reads from the wrap object and shows the frame until the window is closed
        """
        while True:
            if self.show_and_get_frame() is None:
                return

    @abc.abstractmethod
    def _get_frame(self) -> Frame:
        """
        unsafely reads a frame from the wrapped object and returns the read frame
        
        :return: the read frame
        """
        pass

    def show_async(self):
        """
        opens the steam video window on another thread
        """
        Thread(target=self.show).start()



import abc
import pickle
import cv2
import struct

from gbvision.constants.types import Frame
from gbvision.utils.readable import Readable


class StreamReceiver(Readable, abc.ABC):
    """
    this is an abstract receiver that receives stream from a broadcast receiver
    this class should not be instanced but inherited from

    :param shape: optional, the shape (x, y) of the sent frame, when set to something other then (0, 0) it overrides
        the fx and fy parameters, when set to (0, 0) it is not used
    :param fx: ratio between width of the read frame to the width of the frame returned, default is 1 (same width)
    :param fy: ratio between height of the read frame to the height of the frame returned,
        default is 1 (same height)
    """

    def __init__(self, shape=(0, 0), fx: float = 1.0, fy: float = 1.0):
        self.shape = shape
        self.fx = fx
        self.fy = fy
        self.data = b''
        self.payload_size = struct.calcsize("I")

    @abc.abstractmethod
    def _receive(self) -> bytes:
        """
        reads bytes from the stream and returns them
        the amount of bytes read is the choice of the programmer
        for UDP / RAW formats, a large amount is recommended
        for TCP / TCP like formats, a small amount is recommended

        """
        pass

    def _get_frame_data(self) -> bytes:
        """

        :return:
        """
        while len(self.data) < self.payload_size:
            self.data += self._receive()
        packed_msg_size = self.data[:self.payload_size]
        self.data = self.data[self.payload_size:]
        msg_size = struct.unpack("I", packed_msg_size)[0]
        while len(self.data) < msg_size:
            self.data += self._receive()
        frame_data = self.data[:msg_size]
        self.data = self.data[msg_size:]
        return frame_data

    def read(self):
        frame_data = self._get_frame_data()
        frame = pickle.loads(frame_data)
        if frame is None:
            return False, None
        frame = cv2.imdecode(frame, -1)
        return True, self._prep_frame(frame)

    def _prep_frame(self, frame: Frame) -> Frame:
        """
        prepares the frame to be returned and used
        resize and convert to bgr channeled image
        
        :param frame: the frame to be prepared
        :return: the frame after preparations
        """
        if frame is None:
            return frame
        if len(frame.shape) < 3:
            frame = cv2.cvtColor(frame, cv2.COLOR_GRAY2BGR)
        frame = cv2.resize(frame, self.shape, fx=self.fx, fy=self.fy)
        return frame



from setuptools import setup

setup(
    name='gbvision',
    version='0.7.6',
    description='A Python Vision Library for object tracking in the 3D physical space',
    license='Apache License 2.0',
    packages=['gbvision',
              'gbvision/utils',
              'gbvision/utils/net',
              'gbvision/constants',
              'gbvision/exceptions',
              'gbvision/utils/finders',
              'gbvision/gui',
              'gbvision/models',
              'gbvision/tools',
              'gbvision/utils/continuity',
              'gbvision/utils/cameras',
              'gbvision/utils/thresholds',
              'gbvision/utils/shapes',
              'gbvision/utils/recorders',
              'gbvision/utils/denoising'],
    author='Ido Heinemann',
    author_email='idohaineman@gmail.com',
    keywords=['computer vision', 'frc', 'first', 'image processing'],
    url='https://github.com/GreenBlitz/GBVision',
    download_url='https://github.com/GreenBlitz/GBVision/archive/v0.7.6-beta.tar.gz',
    install_requires=[
        'numpy',
        'opencv-python',
        'opencv-contrib-python',
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package

        'Intended Audience :: Developers',  # Define that your audience are developers

        'Programming Language :: Python :: 3',  # Specify which python versions that you want to support
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
)



import abc
from typing import List, Iterable

from gbvision.utils.cameras.camera import Camera

from gbvision.constants.types import Frame, Location, Number, Point, Shape
from gbvision.utils.game_object import GameObject
from gbvision.utils.shapes.base_shape import BaseShapeType


class ObjectFinder(abc.ABC):
    """
    this is an abstract class that represents an object finder
    an object finder is a type that outputs an object's 3d real location based on an of it image it's
    GameObject real-life parameters

    :param game_object: the game object descriptor for the real-life parameters of the finder's target
    :param area_scalar: a scalar to multiply the root of the area of the shape in the image by, default is 1
    """

    def __init__(self, game_object: GameObject, area_scalar=1.0):
        self.game_object = game_object
        self.area_scalar = area_scalar

    def __call__(self, frame: Frame, camera: Camera) -> List[Location]:
        """
        finds all instances of the object in the frame

        :param frame: the frame in which to find
        :param camera: the camera used to read the frame
        :return: all object of this type in the physical space
        """
        return self.locations_from_shapes(self.find_shapes(frame), camera)

    @staticmethod
    @abc.abstractmethod
    def _base_shape() -> BaseShapeType:
        """
        returns the base shape matching this finder

        :return: the base shape (a class that inherits from BaseShape)
        """

    @abc.abstractmethod
    def find_shapes_unsorted(self, frame: Frame) -> List[Shape]:
        """
        finds all the objects and returns them in frame after full pipeline (not sorted)

        :param: The current frame the finder searches in
        :return: A list of objects: see gbvision/constants/types
        """

    def find_shapes(self, frame: Frame) -> List[Shape]:
        """
        finds all the objects and returns them in frame after full pipeline (sorted)

        :param: The current frame the finder searches in
        :return: A list of objects: see gbvision/constants/types
        """
        return self._base_shape().sort_shapes(self.find_shapes_unsorted(frame))

    @classmethod
    def _shape_root_area(cls, shape: Shape) -> Number:
        return cls._base_shape().shape_root_area(shape)

    @classmethod
    def _shape_center(cls, shape: Shape) -> Point:
        return cls._base_shape().shape_center(shape)

    def filter_inner_shapes(self, shapes: List[Shape]) -> List[Shape]:
        """
        filters out all inner shapes in the given sorted list of shapes

        :param shapes: a sorted list of shapes
        :return: the list of shapes, without any shape intersecting with a larger shape
        """
        return self._base_shape().filter_inner_shapes(shapes)

    def locations_from_shapes(self, shapes: Iterable[Shape], camera: Camera) -> List[Location]:
        """
        finds the locations of the shapes based on the shape descriptor and camera constants

        :param shapes: a list of the shapes
        :param camera: the camera used to capture the frame that the shapes were found in
        :return: a list of the locations of all the shapes
        """
        return list(
            map(lambda shape: self.game_object.location_by_params(camera,
                                                                  self._shape_root_area(shape) * self.area_scalar,
                                                                  self._shape_center(shape)), shapes))



from typing import List

import cv2

from gbvision.constants.math import EPSILON
from gbvision.constants.system import CONTOURS_INDEX
from gbvision.constants.types import Contour, Polygon, Point, Frame
from gbvision.utils.pipeline import PipeLine


def __mapper(func) -> PipeLine:
    return PipeLine(lambda x: list(map(func, x)))


@PipeLine
def find_contours(frame: Frame) -> List[Contour]:
    """
    finds the contours in a binary frame

    :param frame: the frame (usually after threshold and denoising)
    :return: a list of all the contours in the frame
    """
    # DO NOT CHANGE THE CHAIN_APPROX_NONE
    # You do not know the damages it may cause
    return cv2.findContours(frame, cv2.RETR_TREE, cv2.CHAIN_APPROX_NONE)[CONTOURS_INDEX]


@PipeLine
def sort_contours(cnts: List[Contour]) -> List[Contour]:
    """
    sorts the list of contours by the contour area

    :param cnts: the list of contours
    :return: the list given, sorted by area
    """
    return sorted(cnts, key=lambda x: cv2.contourArea(x), reverse=True)


class FilterContours(PipeLine):
    """
    a pipeline factory that receives a list of contours and filters out all contours with an area less then defined
    to the pipeline

    :param min_area: the minimal area of a contour in order for it to pass the filter
    """
    def __init__(self, min_area: float):
        PipeLine.__init__(self, lambda cnts: filter(lambda c: cv2.contourArea(c) >= min_area, cnts), list)


convex_hull = PipeLine(cv2.convexHull)

convex_hull_multiple = __mapper(convex_hull)


@PipeLine
def contour_center(cnt: Contour) -> Point:
    """
    finds the center of the contour
    :param cnt: a contour object
    :return: the center of the contour object
    """
    m = cv2.moments(cnt)
    return int(m['m10'] / (m['m00'] + EPSILON)), int(m['m01'] / (m['m00'] + EPSILON))


contours_centers = __mapper(contour_center)

# SHAPES

contours_to_rects = __mapper(cv2.boundingRect)


@PipeLine
def sort_rects(rects):
    return list(sorted(rects, key=lambda x: x[2] * x[3], reverse=True))


contours_to_rects_sorted = contours_to_rects + sort_rects

contours_to_circles = __mapper(cv2.minEnclosingCircle)


@PipeLine
def sort_circles(circs):
    return list(sorted(circs, key=lambda x: x[1], reverse=True))


contours_to_circles_sorted = contours_to_circles + sort_circles

contours_to_rotated_rects = __mapper(cv2.minAreaRect)


@PipeLine
def sort_rotated_rects(rects):
    return list(sorted(rects, key=lambda x: x[1][0] * x[1][1], reverse=True))


contours_to_rotated_rects_sorted = contours_to_rotated_rects + sort_rotated_rects


@PipeLine
def contours_to_ellipses(cnts):
    cnts = filter(lambda x: len(x) >= 5, cnts)
    # ellipse must get contours of at least five points
    return list(map(cv2.fitEllipse, cnts))


sort_ellipses = sort_rotated_rects

contours_to_ellipses_sorted = contours_to_ellipses + sort_ellipses


@PipeLine
def contours_to_polygons(cnts):
    """
    performs approxPolyDP algorithm on a list of contours

    :param cnts: the list of contours
    :return: a list of polygons from the contours
    """
    arc_lengts = map(lambda cnt: 0.05 * cv2.arcLength(cnt, True), cnts)
    return list(map(lambda cnt: cv2.approxPolyDP(cnt, next(arc_lengts), True), cnts))


@PipeLine
def fix_contours_shape(cnts: List[Contour]) -> List[Polygon]:
    """
    fixes the contours to a usable shape
    the shape of the contours is a list of tuples of integers/floats, where each tuple is a point
    an example of two rectangles represented with this shape will be:
    [[(0, 0), (0, 2), (1, 2), (1, 0)],
    [(5, 4), (7, 4), (7, 9), (9, 5)]]
    
    :param cnts: the contours / polygons list whose shape should be fixed
    :return: a list of all the contours with the fixed shape
    """
    cnts = map(lambda polydp: map(lambda x: x[0], polydp), cnts)
    return list(map(lambda polydp: list(map(tuple, polydp)), cnts))


sort_polygons = sort_contours

polygon_center = contour_center

polygons_centers = contours_centers



from .erode import Erode
from .dilate import Dilate
from .erode_and_dilate import ErodeAndDilate
from .median_blur import MedianBlur
from .distance_transform_threshold import DistanceTransformThreshold



from typing import List

from gbvision.models.system import EMPTY_PIPELINE
from gbvision.constants.types import Frame, Polygon, FilterFunction
from gbvision.models.contours import FilterContours, find_contours, sort_polygons, contours_to_polygons, \
    convex_hull_multiple
from .object_finder import ObjectFinder
from gbvision.utils.shapes.base_convex_polygon import BaseConvexPolygon


class ConvexPolygonFinder(ObjectFinder):
    """
    finds any generic polygon, not recommended when another finder can be used

    :param area_scalar: optional, a scalar to multiply the area by, for fine tuning of the function's output
    :param contour_min_area: the minimal area of a contour, used for FilterContours, default is 0 (no area limit)
    :param threshold_func: a pipeline (or any sort of function) that returns a binary threshold of the object
     the finder is searching, the object needs to be white and the rest if the image black (doesn't
     have to be perfect)
    :param contours_process: a pipeline to run on the list of contours (optional)
    :param convex_polygons_process: a pipeline to run on the list of convex polygons (optional)
    """

    @staticmethod
    def _base_shape():
        return BaseConvexPolygon

    def __init__(self, threshold_func: FilterFunction, game_object, area_scalar=1.0, contour_min_area=0.0,
                 contours_process=EMPTY_PIPELINE, convex_polygons_process=EMPTY_PIPELINE):
        ObjectFinder.__init__(self, game_object, area_scalar=area_scalar)
        self._full_pipeline = (EMPTY_PIPELINE +
                               threshold_func +
                               find_contours +
                               FilterContours(min_area=contour_min_area) +
                               contours_process +
                               convex_hull_multiple +
                               contours_to_polygons +
                               convex_polygons_process)

    def find_shapes_unsorted(self, frame: Frame) -> List[Polygon]:
        return self._full_pipeline(frame)



from typing import Union, Tuple
import cv2
import numpy as np

from gbvision.utils.pipeline import PipeLine


class Erode(PipeLine):
    """
    creates a pipeline that erodes the image by a kernel of ones
    used mainly for Erode & Dilate denoise filters

    :param ksize: the kernel size, either an integer (meaning an nxn kernel) or a tuple (nxm kernel)
    :param iterations: optional, the amount of Erode iterations to perform, default is 1.
        Note! a large number of iterations will slow down the program
    :return: a pipeline that erodes the given frame
    """

    def __init__(self, ksize: Union[int, Tuple[int, int]], iterations=1):
        if type(ksize) is int:
            ksize = (ksize, ksize)
        PipeLine.__init__(self, lambda frame: cv2.erode(frame, np.ones(ksize), iterations=iterations))



import numpy as np

from gbvision.models.shapes import rotated_rect_collision
from gbvision.constants.types import RotatedRect, Rect
from .continues_shape import ContinuesShape
from gbvision.utils.shapes.base_rotated_rect import BaseRotatedRect


class ContinuesRotatedRect(ContinuesShape):
    """
    An implementation of ContinuesShape to rotated rectangles.
    used to try and check whether two rotated rectangles are indeed the same one.
    """

    @staticmethod
    def _base_shape():
        return BaseRotatedRect

    def __init__(self, shape: RotatedRect, *args, **kwargs):
        ContinuesShape.__init__(self, shape=shape, *args, **kwargs)

    def _shape_collision(self, shape) -> bool:
        return rotated_rect_collision(self._shape, shape)

    @staticmethod
    def _from_bounding_rect(bounding_rect: Rect) -> RotatedRect:
        # assuming the tilting angle is 0
        return ((bounding_rect[0] - bounding_rect[2]) / 2, (bounding_rect[1] + bounding_rect[3] / 2)), \
               bounding_rect[2:4], 0

    @staticmethod
    def _to_bounding_rect(rotated_rect: RotatedRect) -> Rect:
        x = rotated_rect[0][0]
        y = rotated_rect[0][1]
        w = rotated_rect[1][0]
        h = rotated_rect[1][1]
        a = np.deg2rad(rotated_rect[2])

        bound_w = w * np.cos(a) + h * np.sin(a)
        bound_h = h * np.cos(a) + w * np.sin(a)
        bound_x = x - bound_w / 2
        bound_y = y - bound_h / 2
        return bound_x, bound_y, bound_w, bound_h
